// bmp16.h - header file for writing 16 color bitmaps using the standard CGA palette
// by shaktazuki

#include <stdio.h>
#include <stdlib.h>

#ifndef _bmp16h
#define _bmp16h

// color indices/bitmasks
#define BLACK		0x00
#define BLUE		0x01
#define GREEN		0x02
#define CYAN		0x03
#define RED			0x04
#define MAGENTA		0x05
#define BROWN		0x06
#define LTGRAY		0x07
#define DKGRAY		0x08
#define LTBLUE		0x09
#define LTGREEN		0x0A
#define LTCYAN		0x0B
#define LTRED		0x0C
#define LTMAGENTA	0x0D
#define LTYELLOW	0x0E
#define WHITE		0x0F

#pragma pack(1)
typedef struct tagBITMAPFILEHEADER 
{ 
  unsigned short int bfType; // 0x4D42 "MB" (for some reason it gets reversed on write)
  unsigned int   bfSize; // bfOffBits + int((width * height) / 2.0 + .5)
  unsigned short int	bfReserved1; // 0
  unsigned short int    bfReserved2; // 0
  unsigned int   bfOffBits; // sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO)
} BITMAPFILEHEADER, *PBITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER
{
  unsigned int  biSize; // sizeof(BITMAPINFOHEADER)
  long int   biWidth; 
  long int   biHeight; 
  unsigned short int   biPlanes; // 1
  unsigned short int   biBitCount; // 4
  unsigned int  biCompression; // 0
  unsigned int  biSizeImage; // 0
  long int   biXPelsPerMeter; // 0
  long int   biYPelsPerMeter; // 0
  unsigned int  biClrUsed; // 0
  unsigned int  biClrImportant; // 0
} BITMAPINFOHEADER, *PBITMAPINFOHEADER; 
#pragma pack()

typedef struct tagRGBQUAD 
{
  unsigned char    rgbBlue; 
  unsigned char    rgbGreen; 
  unsigned char    rgbRed; 
  unsigned char    rgbReserved; // 0
} RGBQUAD;

typedef struct tagBITMAPINFO 
{ 
  BITMAPINFOHEADER bmiHeader; 
  RGBQUAD          bmiColors[16]; 
} BITMAPINFO, *PBITMAPINFO;  

BITMAPINFO * getbmp16info(int width, int height) // returns the 16 color bmpinfo struct
{
	BITMAPINFO * info = (BITMAPINFO *) malloc(sizeof(BITMAPINFO));
	info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	info->bmiHeader.biWidth = width;
	info->bmiHeader.biHeight = height;
	info->bmiHeader.biPlanes = 1;
	info->bmiHeader.biBitCount = 4;
	info->bmiHeader.biCompression = 0;
	info->bmiHeader.biSizeImage = 0;
	info->bmiHeader.biXPelsPerMeter = 0;
	info->bmiHeader.biYPelsPerMeter = 0;
	info->bmiHeader.biClrUsed = 0;
	info->bmiHeader.biClrImportant = 0;
	
	// color definitions, whee.
	// fill reserved color with 0
	for(int i = 0; i < 16; i++)
	{
		info->bmiColors[i].rgbReserved = 0;
	}
	// black, 0x000000 (0xrrggbb)
	info->bmiColors[0].rgbRed = 0x00;
	info->bmiColors[0].rgbGreen = 0x00;
	info->bmiColors[0].rgbBlue = 0x00;
	// blue 0x0000AA
	info->bmiColors[1].rgbRed = 0x00;
	info->bmiColors[1].rgbGreen = 0x00;
	info->bmiColors[1].rgbBlue = 0xAA;
	// green 0x00AA00
	info->bmiColors[2].rgbRed = 0x00;
	info->bmiColors[2].rgbGreen = 0xAA;
	info->bmiColors[2].rgbBlue = 0x00;
	// cyan 0x00AAAA
	info->bmiColors[3].rgbRed = 0x00;
	info->bmiColors[3].rgbGreen = 0xAA;
	info->bmiColors[3].rgbBlue = 0xAA;
	// red 0xAA0000
	info->bmiColors[4].rgbRed = 0xAA;
	info->bmiColors[4].rgbGreen = 0x00;
	info->bmiColors[4].rgbBlue = 0x00;
	// magenta 0xAA00AA
	info->bmiColors[5].rgbRed = 0xAA;
	info->bmiColors[5].rgbGreen = 0x00;
	info->bmiColors[5].rgbBlue = 0xAA;
	// brown 0xAA5500
	info->bmiColors[6].rgbRed = 0xAA;
	info->bmiColors[6].rgbGreen = 0x55;
	info->bmiColors[6].rgbBlue = 0x00;
	// light grey 0xAAAAAA
	info->bmiColors[7].rgbRed = 0xAA;
	info->bmiColors[7].rgbGreen = 0xAA;
	info->bmiColors[7].rgbBlue = 0xAA;
	// dark grey 0x555555
	info->bmiColors[8].rgbRed = 0x55;
	info->bmiColors[8].rgbGreen = 0x55;
	info->bmiColors[8].rgbBlue = 0x55;
	// bright blue 0x5555FF
	info->bmiColors[9].rgbRed = 0x55;
	info->bmiColors[9].rgbGreen = 0x55;
	info->bmiColors[9].rgbBlue = 0xFF;
	// bright green 0x55FF55
	info->bmiColors[10].rgbRed = 0x55;
	info->bmiColors[10].rgbGreen = 0xFF;
	info->bmiColors[10].rgbBlue = 0x55;
	// bright cyan 0x55FFFF
	info->bmiColors[11].rgbRed = 0x55;
	info->bmiColors[11].rgbGreen = 0xFF;
	info->bmiColors[11].rgbBlue = 0xFF;
	// bright red 0xFF5555
	info->bmiColors[12].rgbRed = 0xFF;
	info->bmiColors[12].rgbGreen = 0x55;
	info->bmiColors[12].rgbBlue = 0x55;
	// bright magenta 0xFF55FF
	info->bmiColors[13].rgbRed = 0xFF;
	info->bmiColors[13].rgbGreen = 0x55;
	info->bmiColors[13].rgbBlue = 0xFF;
	// bright yellow 0xFFFF55
	info->bmiColors[14].rgbRed = 0xFF;
	info->bmiColors[14].rgbGreen = 0xFF;
	info->bmiColors[14].rgbBlue = 0x55;
	// bright white 0xFFFFFF
	info->bmiColors[15].rgbRed = 0xFF;
	info->bmiColors[15].rgbGreen = 0xFF;
	info->bmiColors[15].rgbBlue = 0xFF;
	return info;
}

BITMAPFILEHEADER * getbmp16header(BITMAPINFO * info) // returns the 16 color bmpheader struct
{
	// necessary to account for added padding
	// (image width / 2) rounded up
	unsigned int data_bytes_per_line = unsigned int(unsigned int(info->bmiHeader.biWidth)/2.0 + .5);
	unsigned int padding_bytes_per_line = (sizeof(long) - data_bytes_per_line % sizeof(long)) % sizeof(long);
	unsigned int bytes_per_scanline = data_bytes_per_line + padding_bytes_per_line;

	BITMAPFILEHEADER * head = (BITMAPFILEHEADER *) malloc(sizeof(BITMAPFILEHEADER));
	head->bfType = 0x4D42; // "MB" which on write somehow becomes "BM"
	head->bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO) + bytes_per_scanline * info->bmiHeader.biHeight;
	head->bfReserved1 = 0;
	head->bfReserved2 = 0;
	head->bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO);
	return head;
}

void writebmp16(char * filename, int width, int height, unsigned char * bitdata) //bitdata should contain int((width * height)/2 + .5)bytes, each of which can reference two colors (ie, 0x0001 -> black pixel, blue pixel)
{
	FILE *file = fopen(filename, "wb");
	if(!file)
	{
		printf("File %s could not be opened.", filename);
		return;
	}
	BITMAPINFO * info = getbmp16info(width, height);
	BITMAPFILEHEADER * head = getbmp16header(info);
	fwrite((void *)head, sizeof(BITMAPFILEHEADER), 1, file);
	fwrite((void *)info, sizeof(BITMAPINFO), 1, file);
	free(head);
	free(info);
	
	// this section is necessary because scanlines must contain sizeof(long) bytes, padded with 0 if need be.
	unsigned int data_bytes_per_line = unsigned int(width/2.0 + .5); // width / 2, rounded up.
	unsigned int padding_bytes_per_line = (sizeof(long) - data_bytes_per_line % sizeof(long)) % sizeof(long);
	unsigned int bytes_per_scanline = data_bytes_per_line + padding_bytes_per_line;
	unsigned char padding = 0;
	for(unsigned int i = 0; i < height * data_bytes_per_line; i += data_bytes_per_line)
	{
		fwrite((void *)(bitdata + i), data_bytes_per_line, 1, file);
		if(padding_bytes_per_line) // if we need padding, this will be other than 0
		{
			fwrite((void *)&padding, 1, padding_bytes_per_line, file);
		}	
	}
	
	// cleanup
	fclose(file);
	return;
}
#endif
	